from __future__ import print_function,division
import numpy as np
from scipy import ndimage
from PIL import Image, ImageDraw


def find_nearest(xx,yy,xval,yval):
    xidx = (np.abs(xx-xval)).argmin()
    yidx = (np.abs(yy-yval)).argmin()

    ridx = yidx / xx.shape[1]
    cidx = xidx % xx.shape[1]

    return [ridx,cidx]

def geoverts_2_imgverts(poly,yy,xx):
    """
    FUNCTION:  verts = geoverts_2_imgverts(xx,yy,poly)
    ARGUMENTS: xx - x-coordinate grid for output raster
               yy - y-coordinate grid for output raster
               poly - swath polygon vertices
    KEYWORDS:  N/A
    RETURNS:   vertsOut - the polygon vertices calculated from the swath
    NOTES:     Takes polygon geographic coordinates and converts them into
               image pixel coordinates for creating a mask image.
    """

    vertsOut = []

    # Iterate through each polygon vertex
    for i in range(len(poly)):
        # Extract x-y coordinate from vertex
        xval = poly[i][0]
        yval = poly[i][1]

        # Find closest image index to the x-y coordinate
        xidx = (np.abs(xx-xval)).argmin()
        yidx = (np.abs(yy-yval)).argmin()

        # Convert the 1-d index to 2-d
        ridx = yidx / xx.shape[1]
        cidx = xidx % xx.shape[1]

        vertsOut.append((int(cidx),int(ridx)))

    return vertsOut

def hist_match(source, template):
    """
    Adjust the pixel values of a grayscale image such that its histogram
    matches that of a target image

    Arguments:
    -----------
        source: np.ndarray
            Image to transform; the histogram is computed over the flattened
            array
        template: np.ndarray
            Template image; can have different dimensions to source
    Returns:
    -----------
        matched: np.ndarray
            The transformed output image
    """

    oldshape = source.shape
    source = source.ravel()
    template = template.ravel()

    # get the set of unique pixel values and their corresponding indices and
    # counts
    s_values, bin_idx, s_counts = np.unique(source, return_inverse=True,
                                            return_counts=True)
    t_values, t_counts = np.unique(template, return_counts=True)

    # take the cumsum of the counts and normalize by the number of pixels to
    # get the empirical cumulative distribution functions for the source and
    # template images (maps pixel value --> quantile)
    s_quantiles = np.cumsum(s_counts).astype(np.float64)
    s_quantiles /= s_quantiles[-1]
    t_quantiles = np.cumsum(t_counts).astype(np.float64)
    t_quantiles /= t_quantiles[-1]

    # interpolate linearly to find the pixel values in the template image
    # that correspond most closely to the quantiles in the source image
    interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)

    return interp_t_values[bin_idx].reshape(oldshape)


def clip(toBeClipped, toClipTo, targetCoordslonKey='ILon',latKey='ILat'):
    if len(targetCoords) != 2:
        raise Error('Wrong input for targetCoords argument, expected list of [lat, lon] keys')

    templateLon = toBeClipped[target]
    templateLat = toBeClipped[latKey]

    mask = np.zeros_like(templateLon)

    minx,maxx = toClipTo['Lon'].min(),toClipTo['Lon'].max()
    miny,maxy = toClipTo['Lat'].min(),toClipTo['Lat'].max()

    mask = np.ma.masked_where(templateLon < minx, mask)
    mask = np.ma.masked_where(templateLon > maxx, mask)
    mask = np.ma.masked_where(templateLat < miny, mask)
    mask = np.ma.masked_where(templateLat > maxy, mask)

    return

def mask(linearRing,yCoords,xCoords):

    verts = geoverts_2_imgverts(linearRing,yCoords,xCoords)

    # Create a mask from the convex hull vertices
    img = Image.new('L', (yCoords.shape[1], yCoords.shape[0]), 0)
    ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
    mask = np.array(img)

    return mask



import ee
import numpy as np
import xarray as xr

import rastersmith as rs

from . import geeutils, downscale

class Sentinel1(object):
    def __init__(self,gr,
                 canny_threshold=7,    # threshold for canny edge detection
                 canny_sigma=1,        # sigma value for gaussian filter
                 canny_lt=7,           # lower threshold for canny detection
                 smoothing=100,        # amount of smoothing in meters
                 connected_pixels=200, # maximum size of the neighborhood in pixels
                 edge_length=50,       # minimum length of edges from canny detection
                 smooth_edges=100):


        self.gr = gr
        self.canny_threshold = canny_threshold
        self.canny_sigma = canny_sigma
        self.canny_lt = canny_lt
        self.smoothing = smoothing
        self.connected_pixels = connected_pixels
        self.edge_length = edge_length
        self.smooth_edges = smooth_edges

        return

    def getFloodMap(self,time_start,time_end):

        geom = ee.Geometry.Rectangle([self.gr.west,self.gr.south,self.gr.east,self.gr.north])

        mapResult = geeutils.s1WaterMap(geom,time_start,time_end,self.canny_threshold,
                                        self.canny_sigma,self.canny_lt,self.smoothing,
                                        self.connected_pixels,self.edge_length,
                                        self.smooth_edges)

        return mapResult


class Atms(object):
    def __init__():

        return


class Landsat8(object):
    def __init__():
        return


class Viirs(object):
    def __init__():
        return


class Sentinel2(object):
    def __init__():
        return



import ee
import datetime

try:
    ee.Initialize()
except EEException as e:
    from oauth2client.service_account import ServiceAccountCredentials
    credentials = ee.ServiceAccountCredentials(
    service_account_email='',
    filename='',
    )
    ee.Initialize(credentials)

landShp = ee.FeatureCollection('USDOS/LSIB/2013')


def getTileLayerUrl(ee_image_object):
    map_id = ee.Image(ee_image_object).getMapId()
    tile_url_template = "https://earthengine.googleapis.com/map/{mapid}/{{z}}/{{x}}/{{y}}?token={token}"
    return tile_url_template.format(**map_id)


def otsu_function(histogram):
    counts = ee.Array(ee.Dictionary(histogram).get('histogram'))
    means = ee.Array(ee.Dictionary(histogram).get('bucketMeans'))
    size = means.length().get([0])
    total = counts.reduce(ee.Reducer.sum(), [0]).get([0])
    sums = means.multiply(counts).reduce(ee.Reducer.sum(), [0]).get([0])
    mean = sums.divide(total)
    indices = ee.List.sequence(1, size)
    #Compute between sum of squares, where each mean partitions the data.

    def bss_function(i):
        aCounts = counts.slice(0, 0, i)
        aCount = aCounts.reduce(ee.Reducer.sum(), [0]).get([0])
        aMeans = means.slice(0, 0, i)
        aMean = aMeans.multiply(aCounts).reduce(ee.Reducer.sum(), [0]).get([0]).divide(aCount)
        bCount = total.subtract(aCount)
        bMean = sums.subtract(aCount.multiply(aMean)).divide(bCount)
        return aCount.multiply(aMean.subtract(mean).pow(2)).add(
               bCount.multiply(bMean.subtract(mean).pow(2)))

    bss = indices.map(bss_function)
    output = means.sort(bss).get([-1])
    return output


def s1WaterMap(geom,iniTime,endTime,
               canny_threshold=7,    # threshold for canny edge detection
               canny_sigma=1,        # sigma value for gaussian filter
               canny_lt=7,           # lower threshold for canny detection
               smoothing=100,        # amount of smoothing in meters
               connected_pixels=200, # maximum size of the neighborhood in pixels
               edge_length=50,       # minimum length of edges from canny detection
               smooth_edges=100):

    def spatialSelect(feature):
        test = ee.Algorithms.If(geom.contains(feature),feature,None)
        return ee.Feature(test)

    collection = ee.ImageCollection('COPERNICUS/S1_GRD')\
                 .filterBounds(geom)\
                 .filter(ee.Filter.listContains('transmitterReceiverPolarisation', 'VV'))\
                 .select('VV').filterDate(iniTime, endTime)

    after = collection.mosaic()

    smoothed = after.focal_median(smoothing, 'circle', 'meters')

    canny = ee.Algorithms.CannyEdgeDetector(smoothed,canny_threshold,canny_sigma)

    connected = canny.mask(canny).lt(canny_lt).connectedPixelCount(connected_pixels, True)
    edges = connected.gte(edge_length)

    edgeBuffer = edges.focal_max(smooth_edges, 'square', 'meters');

    imageEdge = smoothed.mask(edges)

    polygon = ee.Geometry.MultiPolygon(
        [[[[94.32174682617188, 22.916025590238387],
           [94.47761535644531, 22.915393135699805],
           [94.4879150390625, 23.02223600607765],
           [94.32723999023438, 23.026027690875438]]],
         [[[94.14939880371094, 23.187706723229915],
           [94.21119689941406, 23.18581317530292],
           [94.20913696289062, 23.252702265171013],
           [94.14665222167969, 23.250178757002324]]]])

    histogram_image = smoothed.mask(edgeBuffer)
    histogram = histogram_image.reduceRegion(ee.Reducer.histogram(255, 2)\
                                .combine('mean', None, True)\
                                .combine('variance', None,True),polygon,10,bestEffort=True)

    threshold = otsu_function(histogram.get('VV_histogram'));

    countries = landShp.filterBounds(geom).map(spatialSelect,True)

    water = smoothed.mask(smoothed.lt(threshold)).clip(countries)

    mapUrl = getTileLayerUrl(water.visualize(palette='#9999ff'))

    return mapUrl



def SurfaceWaterAlgorithm(aoi,images, pcnt_perm, pcnt_temp, water_thresh, ndvi_thresh, hand_mask):

    STD_NAMES   = ['blue2', 'blue', 'green', 'red', 'nir', 'swir1', 'swir2']

    # calculate percentile images
    prcnt_img_perm = images.reduce(ee.Reducer.percentile([float(pcnt_perm)])).rename(STD_NAMES)
    prcnt_img_temp = images.reduce(ee.Reducer.percentile([float(pcnt_temp)])).rename(STD_NAMES)

    # MNDWI
    MNDWI_perm = prcnt_img_perm.normalizedDifference(['green', 'swir1'])
    MNDWI_temp = prcnt_img_temp.normalizedDifference(['green', 'swir1'])

    # water
    water_perm = MNDWI_perm.gt(float(water_thresh))
    water_temp = MNDWI_temp.gt(float(water_thresh))

    # get NDVI masks
    NDVI_perm_pcnt = prcnt_img_perm.normalizedDifference(['nir', 'red'])
    NDVI_temp_pcnt = prcnt_img_temp.normalizedDifference(['nir', 'red'])
    NDVI_mask_perm = NDVI_perm_pcnt.gt(float(ndvi_thresh))
    NDVI_mask_temp = NDVI_temp_pcnt.gt(float(ndvi_thresh))

    # combined NDVI and HAND masks
    full_mask_perm = NDVI_mask_perm.add(hand_mask)
    full_mask_temp = NDVI_mask_temp.add(hand_mask)

    # apply NDVI and HAND masks
    water_perm_masked = water_perm.updateMask(full_mask_perm.Not())
    water_temp_masked = water_temp.updateMask(full_mask_perm.Not())

    # single image with permanent and temporary water
    water_complete = water_perm_masked.add(water_temp_masked).clip(aoi)

    #return water_complete.updateMask(water_complete)
    return water_complete

def getLandsatCollection(aoi,time_start, time_end, month_index=None, climatology=True, defringe=True, cloud_thresh=None):

    # Landsat band names
    LC457_BANDS = ['B1',    'B1',   'B2',    'B3',  'B4',  'B5',    'B7']
    LC8_BANDS   = ['B1',    'B2',   'B3',    'B4',  'B5',  'B6',    'B7']
    STD_NAMES   = ['blue2', 'blue', 'green', 'red', 'nir', 'swir1', 'swir2']


    # filter Landsat collections on bounds and dates
    L4 = ee.ImageCollection('LANDSAT/LT04/C01/T1_TOA').filterBounds(aoi).filterDate(time_start, ee.Date(time_end).advance(1, 'day'))
    L5 = ee.ImageCollection('LANDSAT/LT05/C01/T1_TOA').filterBounds(aoi).filterDate(time_start, ee.Date(time_end).advance(1, 'day'))
    L7 = ee.ImageCollection('LANDSAT/LE07/C01/T1_TOA').filterBounds(aoi).filterDate(time_start, ee.Date(time_end).advance(1, 'day'))
    L8 = ee.ImageCollection('LANDSAT/LC08/C01/T1_TOA').filterBounds(aoi).filterDate(time_start, ee.Date(time_end).advance(1, 'day'))

    # apply cloud masking
    if int(cloud_thresh) >= 0:
        # helper function: cloud busting
        # (https://code.earthengine.google.com/63f075a9e212f6ed4770af44be18a4fe, Ian Housman and Carson Stam)
        def bustClouds(img):
            t = img
            cs = ee.Algorithms.Landsat.simpleCloudScore(img).select('cloud')
            out = img.mask(img.mask().And(cs.lt(ee.Number(int(cloud_thresh)))))
            return out.copyProperties(t)
        # apply cloud busting function
        L4 = L4.map(bustClouds)
        L5 = L5.map(bustClouds)
        L7 = L7.map(bustClouds)
        L8 = L8.map(bustClouds)

    # select bands and rename
    L4 = L4.select(LC457_BANDS, STD_NAMES)
    L5 = L5.select(LC457_BANDS, STD_NAMES)
    L7 = L7.select(LC457_BANDS, STD_NAMES)
    L8 = L8.select(LC8_BANDS, STD_NAMES)

    # apply defringing
    if defringe == 'true':
        # helper function: defringe Landsat 5 and/or 7
        # (https://code.earthengine.google.com/63f075a9e212f6ed4770af44be18a4fe, Bonnie Ruefenacht)
        k = ee.Kernel.fixed(41, 41, \
        [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]])
        fringeCountThreshold = 279  # define number of non null observations for pixel to not be classified as a fringe
        def defringeLandsat(img):
            m   = img.mask().reduce(ee.Reducer.min())
            sum = m.reduceNeighborhood(ee.Reducer.sum(), k, 'kernel')
            sum = sum.gte(fringeCountThreshold)
            img = img.mask(img.mask().And(sum))
            return img
        L5 = L5.map(defringeLandsat)
        L7 = L7.map(defringeLandsat)

    # merge collections
    images = ee.ImageCollection(L4.merge(L5).merge(L7).merge(L8))

    # filter on selected month
    if climatology:
        if month_index != None:
            images = images.filter(ee.Filter.calendarRange(int(month_index), int(month_index), 'month'))
        else:
            raise ValueError('Month needs to be defined to calculate climatology')

    return images

def JRCAlgorithm(geom,startDate, endDate, month=None):

    IMAGE_COLLECTION = ee.ImageCollection('JRC/GSW1_0/MonthlyHistory')

    myjrc = IMAGE_COLLECTION.filterBounds(geom).filterDate(startDate, endDate)

    if month != None:
        myjrc = myjrc.filter(ee.Filter.calendarRange(int(month), int(month), 'month'))

    # calculate total number of observations
    def calcObs(img):
        # observation is img > 0
        obs = img.gt(0);
        return ee.Image(obs).set('system:time_start', img.get('system:time_start'));

    # calculate the number of times water
    def calcWater(img):
        water = img.select('water').eq(2);
        return ee.Image(water).set('system:time_start', img.get('system:time_start'));

    observations = myjrc.map(calcObs)

    water = myjrc.map(calcWater)

    # sum the totals
    totalObs = ee.Image(ee.ImageCollection(observations).sum().toFloat());
    totalWater = ee.Image(ee.ImageCollection(water).sum().toFloat());

    # calculate the percentage of total water
    returnTime = totalWater.divide(totalObs).multiply(100)

    # make a mask
    water = returnTime.gt(75).rename(['water'])
    water = water.updateMask(water)

    return water


def getHistoricalMap(geom,iniTime,endTime,
                  climatology=True,
                  month=None,
                  defringe=True,
                  pcnt_perm=40,
                  pcnt_temp=8,
                  water_thresh=0.35,
                  ndvi_thresh=0.5,
                  hand_thresh=30,
                  cloud_thresh=10,
                  algorithm='SWT'):

    def spatialSelect(feature):
        test = ee.Algorithms.If(geom.contains(feature),feature,None)
        return ee.Feature(test)

    countries = landShp.filterBounds(geom).map(spatialSelect,True)

    if climatology:
        if month == None:
            raise ValueError('Month needs to be defined to calculate climatology')

    if algorithm == 'SWT':
        # get images
        images = getLandsatCollection(geom,iniTime, endTime, climatology, month, defringe, cloud_thresh)

        # Height Above Nearest Drainage (HAND)
        HAND = ee.Image('users/arjenhaag/SERVIR-Mekong/HAND_MERIT')

        # get HAND mask
        HAND_mask = HAND.gt(float(hand_thresh))


        water = SurfaceWaterAlgorithm(geom,images, pcnt_perm, pcnt_temp, water_thresh, ndvi_thresh, HAND_mask).clip(countries)
        waterMap = getTileLayerUrl(water.updateMask(water.eq(2)).visualize(min=0,max=2,palette='#ffffff,#9999ff,#00008b'))

    elif algorithm == 'JRC':
        water = JRCAlgorithm(geom,iniTime,endTime).clip(countries)
        waterMap = getTileLayerUrl(water.visualize(min=0,max=1,bands='water',palette='#ffffff,#00008b'))

    else:
        raise NotImplementedError('Selected algorithm string not available. Options are: "SWT" or "JRC"')

    return waterMap


def _accumulate(ic,today,ndays=1):

    eeDate = ee.Date(today)

    ic_filtered = ic.filterDate(eeDate.advance(-ndays,'day'),eeDate)

    accum_img = ee.Image(ic_filtered.sum())

    return accum_img.updateMask(accum_img.gt(1))

def getPrecipMap(accumulation=1):
    if accumulation not in [1,3,7]:
        raise NotImplementedError('Selected accumulation value is not yet impleted, options are: 1, 3, 7')

    dt = datetime.datetime.utcnow() - datetime.timedelta(1)
    today = dt.strftime('%Y-%m-%d')

    ic = ee.ImageCollection('JAXA/GPM_L3/GSMaP/v6/operational').select(['hourlyPrecipRateGC'])

    ranges = {1:[1,100],3:[1,250],7:[1,500]}
    crange = ranges[accumulation]

    accum = _accumulate(ic,today,accumulation)

    precipMap = getTileLayerUrl(accum.visualize(min=crange[0],max=crange[1],
                                                palette='#000080,#0045ff,#00fbb2,#67d300,#d8ff22,#ffbe0c,#ff0039,#c95df5,#fef8fe'
                                               )
                               )
    return precipMap

def getAdminMap():
    countries = ee.FeatureCollection('USDOS/LSIB_SIMPLE/2017')

    # Create an empty image into which to paint the features, cast to byte.
    empty = ee.Image().byte()

    # Paint all the polygon edges with the same number and width, display.
    outline = empty.paint(
      featureCollection=countries,
      width=2
    )

    return getTileLayerUrl(outline.visualize())



from __future__ import print_function,division
import numpy as np
import pywt
from scipy import ndimage,signal
from matplotlib import colors
from sklearn.decomposition import PCA
from . import io,utils

def _interp(r,g,b,i,method='cubic'):
    code = {'cubic':3,'bilinear':1,'nearest':0}
    R = ndimage.zoom(r,2,order=code[method])
    G = ndimage.zoom(g,2,order=code[method])
    B = ndimage.zoom(b,2,order=code[method])
    I = ndimage.zoom(i,2,order=code[method])
    
    return R,G,B,I

def _rescale(data,maxVal,minVal,stretch=False,inverse=False):
    if stretch:
        if inverse:
            out = ((maxVal-minVal) * (maxVal-data)) / (maxVal-minVal)
        else:
            out = ((maxVal-minVal) * (data-minVal)) / (maxVal-minVal)
    else:
        if inverse:
            out = (maxVal-data) / (maxVal-minVal)
        else:
            out = (data-minVal) / (maxVal-minVal)
            
    return out


def gramSchmidt():
    
    return


def hsv(R,G,B,I,P,weight=0.3,S=None):
    '''Function to apply hue-saturation-intensity pansharpening algorithm
    
    Args:
        R (ndarray): 2-d array representing the red channel of image.
        G (ndarray): 2-d array representing the green channel of image.
        B (ndarray): 2-d array representing the blue channel of image.
        I (ndarray): 2-d array representing the infrared channel of image.
        P (ndarray): 2-d array representing the panchromatic band of image.
        weight (float): Value representing the weight of IR channel in the pan image. 

    Returns:
        panImg (ndarrayd): 3-d array of the pan sharpened RGB image
    '''
    if not (S is None):
        P = utils.hist_match(P,S)
    
    R,G,B,I = _interp(R,G,B,I)
    rgb = np.dstack([R,G,B])
    hsv = colors.rgb_to_hsv(rgb)
    
    data_max, data_min = hsv[:,:,2].max(),hsv[:,:,2].min()
    intensity = _rescale((P - I*weight),data_max,data_min)

    outHsv = np.dstack([hsv[:,:,0],hsv[:,:,1],intensity])

    panImg = colors.hsv_to_rgb(outHsv)
    
    return panImg


def wavelet(R,G,B,I,P):
    
    interped = _interp(R,G,B,I)
    
    multiSpec = []
    panCoeffs = pywt.dwt2(P, 'haar')
    
    for i in range(len(interped)):
        wave = pywt.dwt2(interped[i], 'haar')
        coeffs = (wave[0],panCoeffs[1])
        multiSpec.append(pywt.idwt2(coeffs, 'haar'))
    
    panImg = np.dstack(multiSpec)
    
    return _rescale(panImg,panImg.max(),panImg.min())
    

def brovey(R,G,B,I,P,weights=None,S=None):
    '''Function to apply brovey pansharpening algorithm
    
    Args:
        R (ndarray): 2-d array representing the red channel of image.
        G (ndarray): 2-d array representing the green channel of image.
        B (ndarray): 2-d array representing the blue channel of image.
        I (ndarray): 2-d array representing the infrared channel of image.
        P (ndarray): 2-d array representing the panchromatic band of image.
        weights (list, ndarray): List representing the weights of channels in 
                                 the pan image in order or RGBI. 

    Returns:
        panImg (ndarray): 3-d array of the pan sharpened RGB image
    '''
    if not (S is None):
        P = utils.hist_match(P,S)
        
    R,G,B,I = _interp(R,G,B,I)
    panImg = np.zeros([I.shape[0],I.shape[1],4])
    
    if weights:
        if type(weights) == str:
            try:
                DNF = (P - (weights[3] * I)) / ((weights[0] * R) + (weights[1] * G) + (weights[2] * B))

                panImg[:,:,2] = B * DNF
                panImg[:,:,1] = G * DNF
                panImg[:,:,0] = R * DNF
                panImg[:,:,3] = I * DNF
            except IndexError:
                raise IndexError('weights list has wrong number of elements')
        else:
            raise TypeError('weights keyword requires a list')
    
    else:
        panImg[:,:,0] = _rescale(R / ((B + G + R) * P),R.max(),R.min())
        panImg[:,:,1] = _rescale(G / ((B + G + R) * P),G.max(),G.min())
        panImg[:,:,2] = _rescale(B / ((B + G + R) * P),B.max(),B.min())
        panImg[:,:,3] = _rescale(I / ((B + G + R) * P),I.max(),I.min())
    
    return panImg


def pca(R,G,B,I,P,n_components=4,S=None):
    '''Function to apply principal component analysis pansharpening algorithm
    
    Args:
        R (ndarray): 2-d array representing the red channel of image.
        G (ndarray): 2-d array representing the green channel of image.
        B (ndarray): 2-d array representing the blue channel of image.
        I (ndarray): 2-d array representing the infrared channel of image.
        P (ndarray): 2-d array representing the panchromatic band of image.
        n_components (int): Integer value representing the number of components to solve 

    Returns:
        panImg (ndarray): 3-d array of the pan sharpened RGB image
    ''' 
    if not (S is None):
        P = utils.hist_match(P,S)
   
    R,G,B,I = _interp(R,G,B,I)
    stack = np.vstack([R.ravel(),G.ravel(),B.ravel(),I.ravel()]).T
    
    mu = np.mean(stack, axis=1)
    
    pca_ = PCA(n_components=n_components)
    pca_.fit(stack)
    
    transform = pca_.transform(stack)
    data_min,data_max = transform[:,0].min(),transform[:,0].max()
    
    transform[:,0] = _rescale(P.ravel(),data_max,data_min,stretch=True)

    Xhat = np.dot(transform[:,:n_components], pca_.components_[:n_components,:])
    Xhat += mu[:,np.newaxis]
    
    panImg = Xhat.reshape([P.shape[0],P.shape[1],Xhat.shape[1]])
    
    return _rescale(panImg,panImg.max(),panImg.min())

    
if __name__ == "__main__":
    infile = '/home/ubuntu/hydra/data/VNP09GA_NRT.A2018053.h27v07.001.h5'
    
    viirs = io.readViirs(infile)
    
    try:
        pan = hsv(viirs)
    except:
        raise RuntimeError('Opps...something went wrong with module')


from __future__ import print_function,division
import itertools
import numpy as np
import xarray as xr
import multiprocessing as mp

import rastersmith as rs


class STARFM(object):
    def __init__(self,gr,windowSize=33,A=0.5):
        self.gr = gr

        if windowSize%2 == 0:
            print("Selected window size of {0} is an even number, \
                  forcing window size to be odd number by subtracting 1")
            windowSize = windowSize-1
        self.w = windowSize

        self.center= int(np.around((self.w-1)/2.))
        self.A = A
        self.buff = int((self.w - 1) / 2)

    def calcWeights(self,imgWindow,slaveWindow,logWeight=False):
        weights = imgWindow.copy()
        weights[:,:,:,:,:] = 1

        yPos = np.array(np.where(weights[:,:,0,:,0])[0]).reshape(weights.shape)
        xPos = np.array(np.where(weights[:,:,0,:,0])[1]).reshape(weights.shape)

        dijk = np.sqrt((self.center-xPos)**2 + (self.center-yPos)**2)
        Dijk = 1 + (dijk/self.A)

        Sijk = np.abs(imgWindow-slaveWindow)

        Tijk = self.tDiff
        if Tijk < 1:
            Tijk = 1

        if logWeight:
            Cijk = np.log(Sijk) * np.log(Tijk) * np.log(Dijk)
        else:
            Cijk = Sijk * Tijk * Dijk


        Wijk = (1/Cijk) / np.sum(1/Cijk)

        return Wijk

    def getDownscaled(self,coarse,fine,slave,pos):
        self.yEdges = [pos[0]-self.buff,pos[0]+self.buff]
        self.xEdges = [pos[1]-self.buff,pos[1]+self.buff]
        # print(pos)

        if ((pos[0] < self.center) | (pos[1] < self.center)) |\
           ((self.gr.dims[0] - pos[0] < self.center) | (self.gr.dims[1] - pos[1] < self.center)):

           result = np.nan

        else:
            # M_t0
            cWindow = coarse.isel(dict(lat=slice(self.yEdges[0],self.yEdges[1]),
                                       lon=slice(self.xEdges[0],self.xEdges[1]))
                                 )
            # L_t0
            fWindow = fine.isel(dict(lat=slice(self.yEdges[0],self.yEdges[1]),
                                     lon=slice(self.xEdges[0],self.xEdges[1]))
                               )
            # M_tk
            sWindow = slave.isel(dict(lat=slice(self.yEdges[0],self.yEdges[1]),
                                      lon=slice(self.xEdges[0],self.xEdges[1]))
                                )

            # Wijk calculation
            weights = self.calcWeights(fWindow,sWindow)#.where(clusters==0)

            # predict the
            outWindow = (fWindow.isel(time=0)/np.sum(weights)) + cWindow.isel(time=0) - sWindow.isel(time=0)

            cntrPx = outWindow.isel(dict(lat=self.center,lon=self.center)).values

            try:
                result = cntrPx.astype(np.float)
            except TypeError:
                result = np.nan

        return result


    def apply(self,coarse,fine,slave,parallel=False):
        self.tDiff = np.abs((coarse.attrs['date']-slave.attrs['date']).days)

        outRast = fine.copy()

        iters = [[i,j] for i in range(self.gr.dims[0]) for j in range(self.gr.dims[1])]

        if parallel:
            ncores = mp.cpu_count() - 1
            p = mp.Pool(ncores)
            args = ((coarse,fine,slave,x) for x in iters)
            pred = p.map(self._applyParallel,args)

        else:
            pred = map(lambda x: self.getDownscaled(coarse,fine,slave,x),iters)

        out = np.array(list(pred)).reshape(outRast.shape)
        outRast[:,:,:,:,:] = out

        outRast.coords['time'] = coarse.coords['time']

        return outRast

    def _applyParallel(self,args):
        coarse,fine,slave,x = args
        return self.getDownscaled(coarse,fine,slave,x)


class STRUM(object):
    def __init__():
        print("Algorithm not yet implemented...check back soon")

        return


class USTARFM(object):
    def __init__():
        print("Algorithm not yet implemented...check back soon")

        return

class Bathtub(object):
    def __init__(self,gr,probablistic=False,demStdDev=4.2):
        self.gr = gr

        return

    def selGridByCell(self,i,j,fWater,hand):
        cell = fWaterSel.isel(dict(lat=i,lon=j))
        x = cell.coords['lon'].values
        y = cell.coords['lat'].values
        res = atms.attrs['resolution']
        return hand.sel(dict(lat=slice(y+res,y-res),lon=slice(x-res,x+res)))

    def grids2raster(self,fWater,raster,grids,yRange,xRange):
        remap = raster.copy()
        # print(remap.shape)
        cnt = 0
        # print(grids[0].shape)
        for i in yRange:
            for j in xRange:
                cell = fWater.isel(dict(lat=i,lon=j))
                x = cell.coords['lon'].values
                y = cell.coords['lat'].values
                res = atms.attrs['resolution']
                remap.sel(dict(lat=slice(y+res,y-res),lon=slice(x-res,x+res)))[:,:,:,:,:] = grids[cnt].values
                cnt += 1

        return remap

    def fillGrid(self,grid,fraction):

        water = grid.copy()
        water.values = np.zeros(water.shape)

        elv = range(15)

        results = list(map(lambda x: xr.where(grid==x,x+1,water),elv))
        result = xr.concat(results,dim='z')

        depth = result.sum(dim='z')
        depth = xr.where(depth==0,30,depth)

        simFracs = list(map(lambda x: depth.where(depth<=x).count().values / float(depth.size),elv))
        diff = np.array(list(map(lambda x: simFracs[x] - fraction, elv)))
        absDiff = np.abs(diff)

        maxDepth = absDiff.argmin()
        try:
            minDepth = np.where(diff<=0)[0].argmax()
        except:
            minDepth = 0
        resid = diff[maxDepth]

        simWater = depth <= maxDepth

        dFlat = depth.isel(dict(time=0,band=0))

        if resid > 0:
            rIdx = np.where(dFlat == maxDepth)
            rand = np.zeros(dFlat.shape) + rIdx[0].size

            rand[rIdx[0],rIdx[1]] = np.random.random(rIdx[0].size) * rIdx[0].size

            thresh = rIdx[0].size * (fraction / simFracs[maxDepth])

            residWater = rand<thresh

        else:
            residWater = np.zeros(dFlat.shape)

        residWater = xr.DataArray(residWater,coords={'lat':simWater.coords['lat'].values,
                                                     'lon':simWater.coords['lon'].values},
                                             dims=['lat','lon'])

        final = simWater.astype(np.bool) | residWater.astype(np.bool)
        final = final.expand_dims('z').transpose('lat','lon','z','band','time')

        return final


    def apply(self,fWater,hand):
        fWaterSel = fWater.sel(dict(lat=slice(self.gr.north,self.gr.south),
                                    lon=slice(self.gr.west,self.gr.east)))

        handSel = hand.sel(dict(lat=slice(self.gr.north,self.gr.south),
                                lon=slice(self.gr.west,self.gr.east)))

        xDim = range(len(fWaterSel.coords['lon'].values))
        yDim = range(len(fWaterSel.coords['lat'].values))

        idxs = [[i,j] for i in yDim for j in xDim]

        fracs = map(lambda i: fWaterSel.isel(dict(lat=i[0],lon=i[1])).values.flatten()[0], idxs)
        grids = map(lambda i: self.selGridByCell(i[0],i[1],fWaterSel,handSel),idxs)

        args = zip(*[grids,fracs])
        waterGrids = map(lambda x: fillGrid(x[0],x[1]),args)

        waterMap = self.grids2raster(fWater,hand,waterGrids,yDim,xDim)

        return waterMap



import netrc
import ftplib
import datetime

def viirs(date,h,v,outdir='./',creds=None):
    """Function to download VIIRS NRT data for specified time and tile

    Args:
        date (datetime.datetime): Datetime object specifying which date the data of interest was acquired.
        h (int): horizontal tile grid to fetch
        v (int): vertical tile grid to fetch
        outdir (str, optional): out directory to dump retrieved data to
        default = './' or current working directory
        creds (str, optional): path to .netrc file with NASA EarthData login in credentials
        default = None

    Returns:
        None
    """

    if outdir[-1] != '/':
        outdir = outdir+'/'

    acct = netrc.netrc(creds)
    usr,_,pswrd = acct.hosts['https://urs.earthdata.nasa.gov']

    basename = 'VNP09GA_NRT.A{0}{1:03d}.h{2:02d}v{3:02d}.001.h5'

    yr = date.year
    dt = (date-datetime.datetime(yr,1,1)).days + 1

    url = 'nrt3.modaps.eosdis.nasa.gov'
    directory = '/allData/5000/VNP09GA_NRT/Recent/'

    ftp = ftplib.FTP(url)
    ftp.login(usr,pswrd)

    ftp.cwd(directory)

    filename = basename.format(yr,dt,h,v)
    with open(outdir + filename, 'wb') as f:
       ftp.retrbinary('RETR ' + filename, f.write)

    ftp.close()

    return

def atms(date,outdir='./'creds=None):
    if outdir[-1] != '/':
        outdir = outdir+'/'
        
    
    
    return

if __name__ == "__main__":
    outdir = '/home/ubuntu/hydra/data/'
    creds = '/home/ubuntu/hydra/earthdata.netrc'
    h = 27
    v = 7
    now = datetime.datetime.now() - datetime.timedelta(1)
    fetchNrtViirs(now,h,v,outdir,creds)
